// docusaurus(J): changing button position
import React, { useEffect, useMemo } from 'react'
import clsx from 'clsx'
import {
  Collapsible,
  ThemeClassNames,
  findFirstCategoryLink,
  isActiveSidebarItem,
  isSamePath,
  useCollapsible,
  useDocSidebarItemsExpandedState,
  usePrevious,
  useThemeConfig,
} from '@docusaurus/theme-common'
import Link from '@docusaurus/Link'
import { translate } from '@docusaurus/Translate'
import DocSidebarItems from '@theme/DocSidebarItems'
import useIsBrowser from '@docusaurus/useIsBrowser' // If we navigate to a category and it becomes active, it should automatically
// expand itself

function useAutoExpandActiveCategory({ isActive, collapsed, setCollapsed }) {
  const wasActive = usePrevious(isActive)
  useEffect(() => {
    const justBecameActive = isActive && !wasActive

    if (justBecameActive && collapsed) {
      setCollapsed(false)
    }
  }, [isActive, wasActive, collapsed, setCollapsed])
}
/**
 * When a collapsible category has no link, we still link it to its first child
 * during SSR as a temporary fallback. This allows to be able to navigate inside
 * the category even when JS fails to load, is delayed or simply disabled
 * React hydration becomes an optional progressive enhancement
 * see https://github.com/facebookincubator/infima/issues/36#issuecomment-772543188
 * see https://github.com/facebook/docusaurus/issues/3030
 */

function useCategoryHrefWithSSRFallback(item) {
  const isBrowser = useIsBrowser()
  return useMemo(() => {
    if (item.href) {
      return item.href
    } // In these cases, it's not necessary to render a fallback
    // We skip the "findFirstCategoryLink" computation

    if (isBrowser || !item.collapsible) {
      return undefined
    }

    return findFirstCategoryLink(item)
  }, [item, isBrowser])
}

function CollapseButton({ categoryLabel, onClick }) {
  return (
    <button
      aria-label={translate(
        {
          id: 'theme.DocSidebarItem.toggleCollapsedCategoryAriaLabel',
          message: "Toggle the collapsible sidebar category '{label}'",
          description:
            'The ARIA label to toggle the collapsible sidebar category',
        },
        {
          label: categoryLabel,
        }
      )}
      type="button"
      className="clean-btn menu__caret"
      onClick={onClick}
    />
  )
}

export default function DocSidebarItemCategory({
  item,
  onItemClick,
  activePath,
  level,
  index,
  ...props
}) {
  const { items, label, collapsible, className, href } = item
  const hrefWithSSRFallback = useCategoryHrefWithSSRFallback(item)
  const isActive = isActiveSidebarItem(item, activePath)
  const isCurrentPage = isSamePath(href, activePath)
  const { collapsed, setCollapsed } = useCollapsible({
    // Active categories are always initialized as expanded. The default
    // (`item.collapsed`) is only used for non-active categories.
    initialState: () => {
      if (!collapsible) {
        return false
      }

      return isActive ? false : item.collapsed
    },
  })
  useAutoExpandActiveCategory({
    isActive,
    collapsed,
    setCollapsed,
  })
  const { expandedItem, setExpandedItem } = useDocSidebarItemsExpandedState()

  function updateCollapsed(toCollapsed = !collapsed) {
    setExpandedItem(toCollapsed ? null : index)
    setCollapsed(toCollapsed)
  }

  const {
    docs: {
      sidebar: { autoCollapseCategories },
    },
  } = useThemeConfig()
  useEffect(() => {
    if (
      collapsible &&
      expandedItem &&
      expandedItem !== index &&
      autoCollapseCategories
    ) {
      setCollapsed(true)
    }
  }, [collapsible, expandedItem, index, setCollapsed, autoCollapseCategories])
  return (
    <li
      className={clsx(
        ThemeClassNames.docs.docSidebarItemCategory,
        ThemeClassNames.docs.docSidebarItemCategoryLevel(level),
        'menu__list-item',
        {
          'menu__list-item--collapsed': collapsed,
        },
        className
      )}
    >
      <div
        className={clsx('menu__list-item-collapsible', {
          'menu__list-item-collapsible--active': isCurrentPage,
        })}
      >
        {href && collapsible && (
          <CollapseButton
            categoryLabel={label}
            onClick={(e) => {
              e.preventDefault()
              updateCollapsed()
            }}
          />
        )}
        <Link
          className={clsx('menu__link', {
            'menu__link--sublist': collapsible,
            'menu__link--sublist-caret': !href && collapsible,
            'menu__link--active': isActive,
          })}
          onClick={
            collapsible
              ? (e) => {
                  onItemClick?.(item)

                  if (href) {
                    updateCollapsed(false)
                  } else {
                    e.preventDefault()
                    updateCollapsed()
                  }
                }
              : () => {
                  onItemClick?.(item)
                }
          }
          aria-current={isCurrentPage ? 'page' : undefined}
          aria-expanded={collapsible ? !collapsed : undefined}
          href={collapsible ? hrefWithSSRFallback ?? '#' : hrefWithSSRFallback}
          {...props}
        >
          {label}
        </Link>
      </div>

      <Collapsible lazy as="ul" className="menu__list" collapsed={collapsed}>
        <DocSidebarItems
          items={items}
          tabIndex={collapsed ? -1 : 0}
          onItemClick={onItemClick}
          activePath={activePath}
          level={level + 1}
        />
      </Collapsible>
    </li>
  )
}
